from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import DomainItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class EromeCrawler:
    def __init__(self, *, include_id: bool, quiet: bool, SQL_Helper: SQLHelper):
        self.include_id = include_id
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director function for Erome scraping"""
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        if 'a' in url.parts:
            domain_obj = await self.handle_album(session, url)
        else:
            domain_obj = await self.handle_profile(session, url)

        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)

        return domain_obj

    async def handle_album(self, session: ScrapeSession, url: URL):
        """Handler function for erome albums, adds media items to the domain item"""
        try:
            domain_obj = DomainItem("erome", {})
            soup = await session.get_BS4(url)
            title = soup.select_one('div[class="col-sm-12 page-content"] h1').get_text()
            if title is None:
                title = url.name
            elif self.include_id:
                title = title + " - " + url.name
            title = await make_title_safe(title)

            # Images
            for link in soup.select('img[class="img-front lasyload"]'):
                link = URL(link['data-src'])
                try:
                    filename, ext = await get_filename_and_ext(link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(url))
                    continue
                link_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("erome", link_path)
                media = MediaItem(link, url, complete, filename, ext)
                await domain_obj.add_media(title, media)

            # Videos
            for link in soup.select('div[class=media-group] div[class=video-lg] video source'):
                link = URL(link['src'])
                try:
                    filename, ext = await get_filename_and_ext(link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue
                link_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("erome", link_path)
                media = MediaItem(link, url, complete, filename, ext)
                await domain_obj.add_media(title, media)

            url_path = await get_db_path(url)
            await self.SQL_Helper.insert_domain("erome", url_path, domain_obj)
            return domain_obj

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return domain_obj

    async def handle_profile(self, session: ScrapeSession, url: URL):
        """Handler for erome profiles, sends albums to handle_album"""
        try:
            domain_obj = DomainItem("erome", {})
            soup = await session.get_BS4(url)
            title = soup.select_one('h1[class="username"]').get_text()
            if title is None:
                title = url.name
            elif self.include_id:
                title = title + " - " + url.name
            title = await make_title_safe(title)

            for album in soup.select("a[class=album-link]"):
                url = URL(album.get('href'))
                await domain_obj.extend(await self.handle_album(session, url))
            await domain_obj.append_title(title)
            return domain_obj

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return domain_obj



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, check_direct, get_filename_and_ext, \
    get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure, InvalidContentTypeFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class CyberdropCrawler:
    def __init__(self, *, include_id=False, quiet: bool, SQL_Helper: SQLHelper):
        self.include_id = include_id
        self.SQL_Helper = SQL_Helper
        self.quiet = quiet

    async def fetch(self, session: ScrapeSession, url: URL):
        """Cyberdrop scraper"""
        album_obj = AlbumItem("Loose Cyberdrop Files", [])

        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        if await check_direct(url):
            url_path = await get_db_path(url)
            complete = await self.SQL_Helper.check_complete_singular("cyberdrop", url_path)
            filename, ext = await get_filename_and_ext(url.name)
            media = MediaItem(url, url, complete, filename, ext)
            await album_obj.add_media(media)
            await self.SQL_Helper.insert_album("cyberdrop", "", album_obj)
            await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
            return album_obj

        try:
            url_path = await get_db_path(url)
            try:
                soup = await session.get_BS4(url)
            except InvalidContentTypeFailure:
                url_path = await get_db_path(url)
                complete = await self.SQL_Helper.check_complete_singular("cyberdrop", url_path)
                filename, ext = await get_filename_and_ext(url.name)
                media = MediaItem(url, url, complete, filename, ext)
                await album_obj.add_media(media)
                await self.SQL_Helper.insert_album("cyberdrop", "", album_obj)
                await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
                return album_obj

            title = soup.select_one("h1[id=title]").get_text()
            if title is None:
                title = url.name
            elif self.include_id:
                titlep2 = url.name
                titlep2 = [s for s in titlep2 if "." in s][-1]
                title = title + " - " + titlep2
            title = await make_title_safe(title.replace(r"\n", "").strip())
            await album_obj.set_new_title(title)

            links = soup.select('div[class="image-container column"] a')
            for link in links:
                link = URL(link.get('href'))
                try:
                    filename, ext = await get_filename_and_ext(link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue

                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("cyberdrop", url_path)
                media = MediaItem(link, url, complete, filename, ext)
                await album_obj.add_media(media)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return album_obj

        url_path = await get_db_path(url)
        await self.SQL_Helper.insert_album("cyberdrop", url_path, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, check_direct, get_filename_and_ext, \
    get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class LoveFapCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.SQL_Helper = SQL_Helper
        self.quiet = quiet

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for lovefap scraping"""
        album_obj = AlbumItem("Loose LoveFap Files", [])

        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        if await check_direct(url):
            url_path = await get_db_path(url)
            complete = await self.SQL_Helper.check_complete_singular("lovefap", url_path)
            filename, ext = await get_filename_and_ext(url.name)
            media = MediaItem(url, url, complete, filename, ext)
            await album_obj.add_media(media)
            await self.SQL_Helper.insert_album("lovefap", "", album_obj)
            await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
            return album_obj

        try:
            if "video" in url.parts:
                await self.fetch_video(session, url, album_obj)
            else:
                await self.fetch_album(session, url, album_obj)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return album_obj

        url_path = await get_db_path(url)
        await self.SQL_Helper.insert_album("cyberdrop", url_path, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj

    async def fetch_album(self, session: ScrapeSession, url: URL, album_obj: AlbumItem):
        """Gets media_items for albums, and adds them to the Album_obj"""
        url_path = await get_db_path(url)
        soup = await session.get_BS4(url)

        title = soup.select_one('div[class=albums-content-header] span[style*="float: left"]').get_text()
        if title is None:
            title = url.name
        title = await make_title_safe(title)
        await album_obj.set_new_title(title)

        links = soup.select('div[class="file picture"] a')
        links.extend(soup.select('div[class="file picture"] a'))
        for link in links:
            link = link.get('href')
            if link.startswith('/'):
                link = url.host + link
            link = URL(link)
            if "video" in link.parts:
                await self.fetch_video(session, url, album_obj)
            else:
                try:
                    filename, ext = await get_filename_and_ext(link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue

                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("lovefap", url_path)
                media = MediaItem(link, url, complete, filename, ext)
                await album_obj.add_media(media)

    async def fetch_video(self, session: ScrapeSession, url: URL, album_obj: AlbumItem):
        """Gets media_items for video links"""
        soup = await session.get_BS4(url)
        video = soup.select_one("video[id=main-video] source")
        if video:
            link = URL(video.get("src"))
            try:
                filename, ext = await get_filename_and_ext(link.name)
            except NoExtensionFailure:
                logger.debug("Couldn't get extension for %s", str(link))
                return

            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("lovefap", url_path)
            media = MediaItem(link, url, complete, filename, ext)
            await album_obj.add_media(media)


import logging

from myjdapi import myjdapi
from yarl import URL

from cyberdrop_dl.base_functions.base_functions import log
from cyberdrop_dl.base_functions.error_classes import JDownloaderFailure


class JDownloader:
    """Class that handles connecting and passing links to JDownloader"""
    def __init__(self, jdownloader_args: dict, quiet: bool):
        self.jdownloader_device = jdownloader_args['jdownloader_device']
        self.jdownloader_username = jdownloader_args['jdownloader_username']
        self.jdownloader_password = jdownloader_args['jdownloader_password']

        self.jdownloader_enable = jdownloader_args['apply_jdownloader']
        self.jdownloader_agent = None
        self.quiet = quiet

    async def jdownloader_setup(self):
        """Setup function for JDownloader"""
        try:
            if not self.jdownloader_username or not self.jdownloader_password or not self.jdownloader_device:
                raise JDownloaderFailure("jdownloader credentials were not provided.")
            jd = myjdapi.Myjdapi()
            jd.set_app_key("CYBERDROP-DL")
            jd.connect(self.jdownloader_username, self.jdownloader_password)
            self.jdownloader_agent = jd.get_device(self.jdownloader_device)
        except JDownloaderFailure:
            await log("[red]Failed JDownloader setup[/red]", quiet=self.quiet)
            self.jdownloader_enable = False

    async def direct_unsupported_to_jdownloader(self, url: URL, title: str):
        """Sends links to JDownloader"""
        if self.jdownloader_enable:
            if not self.jdownloader_agent:
                await self.jdownloader_setup()
            try:
                if "facebook" in url.host.lower() or "instagram" in url.host.lower():
                    raise JDownloaderFailure("Blacklisted META")

                self.jdownloader_agent.linkgrabber.add_links([{
                    "autostart": False,
                    "links": str(url),
                    "packageName": title if title else "Cyberdrop-DL",
                    "overwritePackagizerRules": True
                    }])

            except JDownloaderFailure as e:
                logging.debug(e)
                await log("[red]Failed to send " + str(url) + " to JDownloader[/red]", quiet=self.quiet)



from __future__ import annotations

import logging
import os
import re
from pathlib import Path

import aiofiles
import psutil as psutil
import rich
from yarl import URL

from cyberdrop_dl.base_functions.data_classes import MediaItem
from cyberdrop_dl.base_functions.error_classes import NoExtensionFailure

FILE_FORMATS = {
    'Images': {
        '.jpg', '.jpeg', '.png', '.gif',
        '.gifv', '.webp', '.jpe', '.svg',
        '.jfif', '.tif', '.tiff', '.jif',
    },
    'Videos': {
        '.mpeg', '.avchd', '.webm', '.mpv',
        '.swf', '.avi', '.m4p', '.wmv',
        '.mp2', '.m4v', '.qt', '.mpe',
        '.mp4', '.flv', '.mov', '.mpg',
        '.ogg', '.mkv', '.mts', '.ts',
        '.f4v'
    },
    'Audio': {
        '.mp3', '.flac', '.wav', '.m4a',
    }
}

logger = logging.getLogger(__name__)
MAX_FILENAME_LENGTH = 100


async def clear() -> None:
    """Clears the terminal screen"""
    os.system('cls' if os.name == 'nt' else 'clear')


async def log(text, quiet=False) -> None:
    """Logs to the output log file and optionally prints to the terminal with given style"""
    logger.debug(text)
    if not quiet:
        rich.print(str(text))


async def regex_links(urls: list) -> list:
    """Regex grab the links from the URLs.txt file"""
    """This allows code blocks or full paragraphs to be copy and pasted into the URLs.txt"""
    yarl_links = []
    for line in urls:
        if line.lstrip().startswith('#'):
            continue

        all_links = [x.group().replace(".md.", ".") for x in re.finditer(
            r"(?:http.*?)(?=($|\n|\r\n|\r|\s|\"|\[/URL]|]\[|\[/img]))", line)]
        for link in all_links:
            yarl_links.append(URL(link))
    return yarl_links


async def check_free_space(required_space: int, download_directory: Path) -> bool:
    """Checks if there is enough free space on the drive to continue operating"""
    free_space = psutil.disk_usage(str(download_directory.parent)).free
    free_space = ((free_space / 1024) / 1024) / 1024
    if required_space > free_space:
        return False
    else:
        return True


async def allowed_filetype(media: MediaItem, block_images: bool, block_video: bool, block_audio: bool, block_other: bool):
    """Checks whether the enclosed file is allowed to be downloaded"""
    ext = media.ext
    if block_images:
        if ext in FILE_FORMATS["Images"]:
            return False
    if block_video:
        if ext in FILE_FORMATS["Videos"]:
            return False
    if block_audio:
        if ext in FILE_FORMATS["Audio"]:
            return False
    if block_other:
        if ext not in FILE_FORMATS["Images"] and ext not in FILE_FORMATS["Videos"] and ext not in FILE_FORMATS["Audio"]:
            return False
    return True


async def purge_dir(dirname) -> None:
    """Purges empty directories"""
    deleted = []
    dir_tree = list(os.walk(dirname, topdown=False))

    for tree_element in dir_tree:
        sub_dir = tree_element[0]
        dir_count = len(os.listdir(sub_dir))
        if dir_count == 0:  # Helps with readability and i've had issues with it deleting non-empty dirs
            deleted.append(sub_dir)
    list(map(os.rmdir, deleted))


async def write_last_post_file(file: Path, url: str):
    """Writes the last post url from a thread to the specified file"""
    async with aiofiles.open(file, mode='a') as f:
        await f.write(url + '\n')
    return


async def get_db_path(url: URL, referer=None):
    """Gets the URL path to be put into the DB and checked from the DB"""
    url_path = url.path
    if 'anonfiles' in url.host or 'bayfiles' in url.host:
        url_path = url.path
        url_path = url_path.split('/')
        url_path.pop(0)
        if len(url_path) > 1:
            url_path.pop(1)
        url_path = '/' + '/'.join(url_path)
    if referer:
        if "e-hentai" in referer:
            url_path = url.path
            url_path = url_path.split('keystamp')[0][:-1]

    return url_path


async def sanitize(name: str) -> str:
    """Simple sanitization to remove illegal characters"""
    return re.sub(r'[<>:"/\\|?*\']', "", name).strip()


async def make_title_safe(title: str) -> str:
    """Simple sanitization to remove illegal characters from titles and trim the length to be less than 60 chars"""
    title = re.sub(r'[\\*?:"<>|./]', "-", title)
    title = title[:60].strip()
    return title


async def check_direct(url: URL):
    """Checks whether the given url is a direct link to a content item"""
    mapping_direct = [r'i.pixl.li', r'i..pixl.li', r'img-...cyberdrop...', r'f.cyberdrop...',
                      r'fs-...cyberdrop...', r'jpg.church/images/...', r'simp..jpg.church', r's..putmega.com',
                      r's..putme.ga', r'images..imgbox.com', 's..lovefap...']
    return any(re.search(domain, url.host) for domain in mapping_direct)


async def get_filename_and_ext(filename, forum=False):
    """Returns the filename and extension of a given file, throws NoExtensionFailure if there is no extension"""
    filename_parts = filename.rsplit('.', 1)
    if len(filename_parts) == 1:
        raise NoExtensionFailure()
    if filename_parts[-1].isnumeric() and forum:
        filename_parts = filename_parts[0].rsplit('-', 1)
    ext = "." + filename_parts[-1].lower()
    filename = await sanitize(filename_parts[0] + ext)
    return filename, ext



from __future__ import annotations

from rich.panel import Panel
from rich.progress import Progress, BarColumn, SpinnerColumn, TransferSpeedColumn, DownloadColumn, TimeRemainingColumn
from rich.table import Table

forum_progress = Progress("[progress.description]{task.description}",
                         BarColumn(bar_width=None),
                         "[progress.percentage]{task.percentage:>3.2f}%",
                         "{task.completed} of {task.total} Threads Completed")

cascade_progress = Progress("[progress.description]{task.description}",
                            BarColumn(bar_width=None),
                            "[progress.percentage]{task.percentage:>3.2f}%",
                            "{task.completed} of {task.total} Domains Completed")

domain_progress = Progress("[progress.description]{task.description}",
                           BarColumn(bar_width=None),
                           "[progress.percentage]{task.percentage:>3.2f}%",
                           "{task.completed} of {task.total} Albums Completed")

album_progress = Progress("[progress.description]{task.description}",
                          BarColumn(bar_width=None),
                          "[progress.percentage]{task.percentage:>3.2f}%",
                          "{task.completed} of {task.total} Files Completed")

overall_file_progress = Progress("[progress.description]{task.description}",
                                 BarColumn(bar_width=None),
                                 "[progress.percentage]{task.percentage:>3.2f}%",
                                 "{task.completed} of {task.total} Files Completed")

file_progress = Progress(SpinnerColumn(),
                         "[progress.description]{task.description}",
                         BarColumn(bar_width=None),
                         "[progress.percentage]{task.percentage:>3.2f}%",
                         "━",
                         DownloadColumn(),
                         "━",
                         TransferSpeedColumn(),
                         "━",
                         TimeRemainingColumn())


async def get_forum_table(progress_options: dict):
    """Table creator for forum threads"""
    progress_table = Table.grid(expand=True)
    if not progress_options['dont_show_overall_progress']:
        progress_table.add_row(Panel.fit(overall_file_progress, title="Overall Progress", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_forum_progress']:
        progress_table.add_row(Panel.fit(forum_progress, title="Total Thread", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_thread_progress']:
        progress_table.add_row(Panel.fit(cascade_progress, title="Current Thread", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_domain_progress']:
        progress_table.add_row(Panel.fit(domain_progress, title="Domains Being Downloaded", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_album_progress']:
        progress_table.add_row(Panel.fit(album_progress, title="Albums Being Downloaded", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_file_progress']:
        progress_table.add_row(Panel.fit(file_progress, title="[b]Files Being Downloaded", border_style="green", padding=(1, 1)))
    return progress_table


async def get_cascade_table(progress_options: dict):
    """Table creator for cascade objects"""
    progress_table = Table.grid(expand=True)
    if not progress_options['dont_show_overall_progress']:
        progress_table.add_row(Panel.fit(overall_file_progress, title="Overall Progress", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_thread_progress']:
        progress_table.add_row(Panel.fit(cascade_progress, title="Current Thread", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_domain_progress']:
        progress_table.add_row(Panel.fit(domain_progress, title="Domains Being Downloaded", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_album_progress']:
        progress_table.add_row(Panel.fit(album_progress, title="Albums Being Downloaded", border_style="green", padding=(1, 1)))
    if not progress_options['dont_show_file_progress']:
        progress_table.add_row(Panel.fit(file_progress, title="[b]Files Being Downloaded", border_style="green", padding=(1, 1)))
    return progress_table



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import ClassVar, List, Tuple

from yarl import URL


@dataclass
class MediaItem:
    url: URL
    referer: URL
    complete: bool
    filename: str
    ext: str

    async def is_complete(self):
        return self.complete

    async def mark_completed(self):
        self.complete = True


@dataclass
class AlbumItem:
    """Class for keeping track of download links for each album"""
    title: str
    media: List[MediaItem]

    async def add_media(self, media_item: MediaItem):
        self.media.append(media_item)

    async def set_new_title(self, new_title: str):
        self.title = new_title

    async def append_title(self, title):
        new_title = title + '/' + self.title
        self.title = new_title

    async def extend(self, album):
        self.media.extend(album.media)

    async def is_empty(self):
        if not self.media:
            return True
        return False


@dataclass
class DomainItem:
    """Class for keeping track of albums for each scraper type"""
    domain: str
    albums: dict

    async def add_to_album(self, title: str, media: MediaItem):
        if title in self.albums.keys():
            await self.albums[title].add_media(media)
        else:
            self.albums[title] = AlbumItem(title=title, media=[media])

    async def add_media(self, title: str, media: MediaItem):
        if title in self.albums.keys():
            album = self.albums[title]
            await album.add_media(media)
        else:
            self.albums[title] = AlbumItem(title, [media])

    async def add_album(self, title: str, album: AlbumItem):
        if title in self.albums.keys():
            stored_album = self.albums[title]
            for media_item in album.media:
                if media_item in stored_album.media:
                    continue
                await stored_album.add_media(media_item)
        else:
            self.albums[title] = album

    async def set_new_domain(self, domain: str):
        self.domain = domain

    async def extend(self, domain):
        for title, album in domain.albums.items():
            await self.add_album(title, album)

    async def append_title(self, title):
        if not title:
            return
        new_albums = {}
        for album_str, album in self.albums.items():
            new_title = title+'/'+album_str
            new_albums[new_title] = album
            album.title = new_title
        self.albums = new_albums


@dataclass
class CascadeItem:
    """Class for keeping track of domains for each scraper type"""
    domains: dict

    async def add_albums(self, domain_item: DomainItem):
        domain = domain_item.domain
        albums = domain_item.albums
        for title, album in albums.items():
            await self.add_album(domain, title, album)

    async def add_to_album(self, domain: str, title: str, media_item: MediaItem):
        if domain in self.domains.keys():
            await self.domains[domain].add_to_album(title, media_item)
        else:
            self.domains[domain] = DomainItem(domain, {title: AlbumItem(title, [media_item])})

    async def add_album(self, domain: str, title: str, album: AlbumItem):
        if domain in self.domains.keys():
            await self.domains[domain].add_album(title, album)
        else:
            self.domains[domain] = DomainItem(domain, {title: album})

    async def is_empty(self):
        for _, domain in self.domains.items():
            for _, album in domain.albums.items():
                if album.media:
                    return False
        return True

    async def get_total(self):
        total = 0
        for _, domain in self.domains.items():
            for _, album in domain.albums.items():
                total += len(album.media)
        return total

    async def append_title(self, title: str):
        if not title:
            return
        for _, domain in self.domains.items():
            new_albums = {}
            for album_str, album in domain.albums.items():
                new_title = title+'/'+album_str
                new_albums[new_title] = album
                album.title = new_title
            domain.albums = new_albums

    async def extend(self, Cascade):
        if Cascade:
            if Cascade.domains:
                for domain_str, domain in Cascade.domains.items():
                    for album_str, album in domain.albums.items():
                        await self.add_album(domain_str, album_str, album)

    async def dedupe(self):
        for _, domain in self.domains.items():
            for _, album in domain.albums.items():
                check = []
                allowed = []
                for media_item in album.media:
                    if media_item.url in check:
                        continue
                    check.append(media_item.url)
                    allowed.append(media_item)
                album.media = allowed


@dataclass
class ForumItem:
    """Class for keeping track of forum threads"""
    threads: dict

    async def add_album_to_thread(self, title: str, domain: str, album: AlbumItem):
        if title not in self.threads.keys():
            self.threads[title] = CascadeItem({domain: DomainItem(domain, {album.title: album})})
        else:
            await self.threads[title].add_album(domain, album.title, album)

    async def add_thread(self, title: str, cascade: CascadeItem):
        if title not in self.threads.keys():
            self.threads[title] = cascade
        else:
            await self.threads[title].extend(cascade)

    async def is_empty(self):
        for _, Cascade in self.threads.items():
            for _, domain in Cascade.domains.items():
                for _, album in domain.albums.items():
                    if album.media:
                        return False
        return True

    async def get_total(self):
        total = 0
        for _, Cascade in self.threads.items():
            for _, domain in Cascade.domains.items():
                for _, album in domain.albums.items():
                    total += len(album.media)
        return total

    async def dedupe(self):
        for _, Cascade in self.threads.items():
            for _, domain in Cascade.domains.items():
                for _, album in domain.albums.items():
                    check = []
                    allowed = []
                    for media_item in album.media:
                        if media_item.url in check:
                            continue
                        check.append(media_item.url)
                        allowed.append(media_item)
                    album.media = allowed

    async def extend_thread(self, title: str, cascade: CascadeItem):
        if title in self.threads.keys():
            await self.threads[title].extend(cascade)
        else:
            self.threads[title] = cascade

@dataclass
class FileLock:
    """Rudimentary file lock system"""
    locked_files = []

    async def check_lock(self, filename):
        await asyncio.sleep(.1)
        return filename.lower() in self.locked_files

    async def add_lock(self, filename):
        self.locked_files.append(filename.lower())

    async def remove_lock(self, filename):
        self.locked_files.remove(filename.lower())


@dataclass
class SkipData:
    """The allows optoins for domains to skip when scraping"""
    supported_hosts: ClassVar[Tuple[str]] = ("anonfiles", "bayfiles", "bunkr", "coomer.party", "cyberdrop",
                                             "cyberfile", "e-hentai", "erome", "fapello", "gfycat", "gofile",
                                             "hgamecg", "img.kiwi", "imgbox", "jpg.church", "jpg.fish", "kemono.party",
                                             "lovefap", "nsfw.xxx", "pimpandhost", "pixeldrain", "pixl.li", "postimg",
                                             "saint", "simpcity", "socialmediagirls", "xbunker", "xbunkr")
    sites: List[str]



from yarl import URL

from ..base_functions.base_functions import log, logger, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class GfycatCrawler:
    def __init__(self, *, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Basic scraper for gfycat"""
        try:
            album_obj = AlbumItem("Gfycat", [])
            await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
            soup = await session.get_BS4(url)

            video = soup.select_one("video[class='video media']")
            video_srcs = video.select("source")

            video_link = None
            for src in video_srcs:
                link = URL(src.get("src"))
                if "giant" in link.host:
                    video_link = link
                    break
            if video_link is None:
                video_link = URL(video_srcs[0].get("src"))

            url_path = await get_db_path(video_link)
            complete = await self.SQL_Helper.check_complete_singular("gfycat", url_path)
            filename, ext = await get_filename_and_ext(video_link.name)
            media_item = MediaItem(video_link, url, complete, filename, ext)
            await album_obj.add_media(media_item)
            await self.SQL_Helper.insert_album("gfycat", url_path, album_obj)
            await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
            return album_obj

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return album_obj



from yarl import URL

from ..base_functions.base_functions import log, logger, get_db_path, get_filename_and_ext
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class PostImgCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for PostImg scraping"""
        album_obj = AlbumItem("Loose PostImg Files", [])
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        try:
            if "gallery" in url.parts:
                content = await self.get_folder(session, url)
                for media_item in content:
                    await album_obj.add_media(media_item)
            else:
                media_item = await self.get_singular(session, url)
                await album_obj.add_media(media_item)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        await self.SQL_Helper.insert_album("postimg", url.path, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj

    async def get_folder(self, session: ScrapeSession, url: URL):
        """Handles folder scraping"""
        album = url.raw_name
        data = {"action": "list", "album": album}
        content = []
        i = 1
        while True:
            data_used = data
            data_used["page"] = i
            data_out = await session.post(URL("https://postimg.cc/json"), data_used)
            if data_out['status_code'] != 200 or not data_out['images']:
                break
            for item in data_out['images']:
                referer = URL("https://postimg.cc/" + item[0])
                img = URL(item[4].replace(item[0], item[1]))

                try:
                    filename, ext = await get_filename_and_ext(img.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(img))
                    continue
                url_path = await get_db_path(img)
                complete = await self.SQL_Helper.check_complete_singular("postimg", url_path)
                media_item = MediaItem(img, referer, complete, filename, ext)

                content.append(media_item)
            i += 1
        return content

    async def get_singular(self, session: ScrapeSession, url: URL):
        """Handles singular folder scraping"""
        soup = await session.get_BS4(url)
        link = URL(soup.select_one("a[id=download]").get('href').replace("?dl=1", ""))

        filename, ext = await get_filename_and_ext(link.name)
        url_path = await get_db_path(link)
        complete = await self.SQL_Helper.check_complete_singular("postimg", url_path)
        media_item = MediaItem(link, url, complete, filename, ext)

        return media_item
